/*
 * Quake benchmark
 * Loukas Kallivokas and David O'Hallaron
 * Carnegie Mellon University, November, 1997
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mpi.h"
#include <sys/time.h>

#define AMAX_NAME  128

#ifndef PI
#  define PI 3.141592653589793238
#endif

double rtclock()
{ 
  struct timezone Tzp; 
  struct timeval Tp; 
  int stat; 
  stat = gettimeofday (&Tp, &Tzp); 
  if (stat != 0) printf("Error return from gettimeofday: %d",stat); 
  return(Tp.tv_sec + Tp.tv_usec*1.0e-6); 
}


struct options {
    int quiet;    /* run quietly unless there are errors (-Q) */
    int help;     /* do we want to print a help message (-h -H) */
};

struct excitation {

  double dt;       /* time step */
  double duration; /* total duration */
  double t0;       /* rise time */

};

struct damping {

  double zeta, consta, constb, freq;

};

struct properties {

  double cp;        /* compressional wave velocity */
  double cs;        /* shear wave velocity */
  double den;       /* density */

};

struct source {

  double dip, strike, rake, fault;
  double xyz[3];
  double epixyz[3];
  int sourcenode;
  int epicenternode;

};


/* name of the main program */

char *progname;

struct options options;

/* global packfile variables */

FILE *packfile;

/* global Archimedes variables */

int ARCHnodes;
int *ARCHstart;
int *ARCHend; /* one more than the index of last element */
int *ARCHsize;
int ARCHpriv;
int ARCHmine;
int ARCHelems;
int ARCHglobalnodes;
int ARCHmesh_dim;
int ARCHglobalelems;
int ARCHcorners;
int ARCHsubdomains;
double ARCHduration;
int ARCHmatrixlen;
int ARCHcholeskylen;

int *ARCHglobalnode;
int *ARCHglobalelem;
double (*ARCHcoord)[3];
int (*ARCHvertex)[4];
int *ARCHmatrixcol;
int *ARCHmatrixindex;

int numProcs, myRank;
int *numGhosts;
int totalGhosts;
int *ghostStart;
int **globalGhostIndices;

int *numRemoteGhosts;
int **remoteGhostIndices;
double **remoteGhostBuffer;
FILE *dbgFile;
char fname[30];
int sourcenodeRank;
int epicenternodeRank;

/* functions */

void arch_init(int argc, char **argv, struct options *op);
void mem_init(void);
void mem_init_basic(void);
void mem_init_K(void);
void arch_readnodevector(double *v, int n);
void slip(double *u, double *v, double *w);
double distance(double p1[], double p2[]);
void centroid(double x[][3], double xc[]);
double point2fault(double x[]);
void abe_matrix(double vertices[][3], int bv[],
		struct properties *prop, double Ce[]);
void element_matrices(double vertices[][3], struct properties *prop,
		      double Ke[][12], double Me[]);
void vv12x12(double v1[], double v2[], double u[]);
void mv12x12(double m[][12], double v[]);
void smvp(int nodes, double ***A, int *Acol, int *Aindex,
		double **v, double **w);
void smvp_opt(int nodes, double ***A, int *Acol, int *Aindex,
		double **v, double **w);

double phi0(double t);
double phi1(double t);
double phi2(double t);

void distribute_arrays();
void add_ghosts_to_disp( int which_disp );
void update_disp_ghosts( int which_disp );

/* global simulation variables */

int *nodekind;
double *nodekindf;
int *source_elms;
double (*M)[3],(*C)[3],(*M23)[3],(*C23)[3],(*V23)[3], (*vel)[3];
double (*disp0)[3],(*disp1)[3],(*disp2)[3], (*K)[9];

struct source Src;
struct excitation Exc;
struct damping Damp;

/*--------------------------------------------------------------------------*/ 

int main(int argc, char **argv)
{
  int i, j, k, ii, jj, kk, iter, timesteps;
  int disptplus, dispt, disptminus;
  int verticesonbnd;
  int cor[4], bv[4];
  int Step_stride;

  double time;
  double Ke[12][12], Me[12], Ce[12], Mexv[12], Cexv[12], v[12];
  double alpha, c0[3], d1, d2, bigdist1, bigdist2, xc[3], uf[3];
  double vertices[4][3];

  struct properties prop;

/* NOTE: There are 5 possible flag values for the node data:

           1 if the node is in the interior 
           4 if the node is on a x=const boundary surface
           5 if the node is on a y=const boundary surface
           6 if the node is on the bottom surface (z=z_lower)
           3 if the node is on the surface (z=0) (but not along the edges)
*/

/*--------------------------------------------------------------------------*/ 

  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
  MPI_Comm_size(MPI_COMM_WORLD, &numProcs);


  if ( myRank == 0 ) {

    printf("Calling arch_init\n");
    /* Read in data from the pack file */  
    arch_init(argc, argv, &options);


    /* Dynamic memory allocations and initializations */
    mem_init();

    arch_readnodevector(nodekindf, ARCHnodes);

    if (!options.quiet)
      fprintf(stderr, "%s: Beginning simulation.\n", argv[0]);

    /* Excitation characteristics */

    Exc.dt = 0.0024;
    Exc.duration = ARCHduration;
    Exc.t0 = 0.6;
    timesteps = Exc.duration / Exc.dt + 1;

    /* Damping characteristics */

    Damp.zeta = 30.0;
    Damp.consta = 0.00533333;
    Damp.constb = 0.06666667;
    Damp.freq = 0.5;

    /* Source characteristics */

    Src.strike = 111.0 * PI / 180.0;
    Src.dip = 44.0 * PI / 180.0;
    Src.rake = 70.0 * PI / 180.0;
    Src.fault = 29.640788;
    Src.xyz[0] = 32.264153;
    Src.xyz[1] = 23.814432;
    Src.xyz[2] = - 11.25;
    Src.epixyz[0] = Src.xyz[0];
    Src.epixyz[1] = Src.xyz[1];
    Src.epixyz[2] = 0.0;
    Src.sourcenode = - 1;
    Src.epicenternode = - 1;

    /* Prescribe slip motion */

    uf[0] = uf[1] = uf[2] = 0.0;
    slip(&uf[0], &uf[1], &uf[2]);
    uf[0] *= Src.fault;
    uf[1] *= Src.fault;
    uf[2] *= Src.fault;

    /* Soil properties (homogeneous material) */

    prop.cp = 6.0;
    prop.cs = 3.2;
    prop.den = 2.0;

    /* Output frequency parameter */

    Step_stride = 30;

    disptplus = 0;
    dispt = 1;
    disptminus = 2;

    /* Case info */

    fprintf(stderr, "\n");
    fprintf(stderr, "CASE SUMMARY\n");
    fprintf(stderr, "Fault information\n");
    fprintf(stderr, "  Orientation:  strike: %f\n", Src.strike);
    fprintf(stderr, "                   dip: %f\n", Src.dip);
    fprintf(stderr, "                  rake: %f\n", Src.rake);
    fprintf(stderr, "           dislocation: %f cm\n", Src.fault);
    fprintf(stderr, "Hypocenter: (%f, %f, %f) Km\n", 
	    Src.xyz[0], Src.xyz[1], Src.xyz[2]);
    fprintf(stderr, "Excitation characteristics\n");
    fprintf(stderr, "     Time step: %f sec\n", Exc.dt);
    fprintf(stderr, "      Duration: %f sec\n", Exc.duration);
    fprintf(stderr, "     Rise time: %f sec\n", Exc.t0);
    fprintf(stderr, "\n");
    fflush(stderr);

    /* Redefine nodekind to be 1 for all surface nodes */

    for (i = 0; i < ARCHnodes; i++) {
      nodekind[i] = (int) nodekindf[i];
      if (nodekind[i] == 3) 
	nodekind[i] = 1;
    }

    /* Search for the node closest to the point source (hypocenter) and */
    /*        for the node closest to the epicenter */

    bigdist1 = 1000000.0;
    bigdist2 = 1000000.0;

    for (i = 0; i < ARCHnodes; i++) {
      c0[0] = ARCHcoord[i][0];
      c0[1] = ARCHcoord[i][1];
      c0[2] = ARCHcoord[i][2];
      d1 = distance(c0, Src.xyz);
      d2 = distance(c0, Src.epixyz);

      if (d1 < bigdist1) {
	bigdist1 = d1;
	Src.sourcenode = i;
      }

      if (d2 < bigdist2) {
	bigdist2 = d2;
	Src.epicenternode = i;
      }

    }

    if (Src.sourcenode != 0 && Src.sourcenode <= ARCHmine) {
      fprintf(stderr, "The source is node %d at (%f  %f  %f)\n", 
	      ARCHglobalnode[Src.sourcenode], 
	      ARCHcoord[Src.sourcenode][0],
	      ARCHcoord[Src.sourcenode][1],
	      ARCHcoord[Src.sourcenode][2]);
      fflush(stderr);
    }

    if (Src.epicenternode != 0 && Src.epicenternode <= ARCHmine) {
      fprintf(stderr, "The epicenter is node %d at (%f  %f  %f)\n", 
	      ARCHglobalnode[Src.epicenternode],
	      ARCHcoord[Src.epicenternode][0],
	      ARCHcoord[Src.epicenternode][1],
	      ARCHcoord[Src.epicenternode][2]);
      fflush(stderr);
    }

    /* Search for all the elements that contain the source node */

    if (Src.sourcenode != 0) {

      for (i = 0; i < ARCHelems; i++) {
	for (j = 0; j < 4; j++)
	  cor[j] = ARCHvertex[i][j];

	if (cor[0] == Src.sourcenode || cor[1] == Src.sourcenode ||
	    cor[2] == Src.sourcenode || cor[3] == Src.sourcenode) {

	  for (j = 0; j < 4; j++)
	    for (k = 0; k < 3; k++)
	      vertices[j][k] = ARCHcoord[cor[j]][k];

	  centroid(vertices, xc);

	  source_elms[i] = 2;
	  if (point2fault(xc) >= 0) 
	    source_elms[i] = 3;

	}
      }
    }

    /* Simulation */

    for (i = 0; i < ARCHelems; i++) {
      for (j = 0; j < 12; j++) {
	Me[j] = 0.0;
	Ce[j] = 0.0;
	v[j] = 0.0;
	for (k = 0; k < 12; k++) 
	  Ke[j][k] = 0.0;
      }

      for (j = 0; j < 4; j++) {
	cor[j] = ARCHvertex[i][j];
      }

      verticesonbnd = 0;
      for (j = 0; j < 4; j++)
	if (nodekind[cor[j]] != 1)
	  bv[verticesonbnd++] = j;

      /*
	if (verticesonbnd == 4) {
	fprintf (stderr, "Warning! 4 vertices seem to be on the boundary\n");
	for (j = 0; j < 4; j++)
	fprintf (stderr, "%f %f %f nodekind[cor[%d]]=%d\n",
	ARCHcoord[cor[bv[j]]][0],
	ARCHcoord[cor[bv[j]]][1],
	ARCHcoord[cor[bv[j]]][2],j,nodekind[cor[j]]);
	}
      */

      if (verticesonbnd == 3) {
	for (j = 0; j < 3; j++)
	  for (k = 0; k < 3; k++)
	    vertices[j][k] = ARCHcoord[cor[bv[j]]][k];

	abe_matrix(vertices, bv, &prop, Ce);

      }

      for (j = 0; j < 4; j++)
	for (k = 0; k < 3; k++)
	  vertices[j][k] = ARCHcoord[cor[j]][k];

      element_matrices(vertices, &prop, Ke, Me);

      /* Damping (proportional) */

      centroid(vertices, xc);

      if (xc[2] < - 11.5)
	alpha = 2.0 * Damp.zeta / 100.0 * (2.0 * PI * Damp.freq);
      else
	alpha = 4.0 * PI * Damp.consta * 0.95 / (prop.cs + Damp.constb);

      for (j = 0; j < 12; j++)
	Ce[j] = Ce[j] + alpha * Me[j];

      /* Source mechanism */

      if (source_elms[i] == 2 || source_elms[i] == 3) {

	for (j = 0; j < 4; j++) {

	  if (cor[j] == Src.sourcenode) {

	    v[3 * j] = uf[0];
	    v[3 * j + 1] = uf[1];
	    v[3 * j + 2] = uf[2];

	  } else {

	    v[3 * j] = 0;
	    v[3 * j + 1] = 0;
	    v[3 * j + 2] = 0;

	  }
	}

	vv12x12(Me, v, Mexv);
	vv12x12(Ce, v, Cexv);
	mv12x12(Ke, v);

	if (source_elms[i] == 3)
	  for (j = 0; j < 12; j++) {
	    v[j] = - v[j];
	    Mexv[j] = - Mexv[j];
	    Cexv[j] = - Cexv[j];
	  }

	/* Assemble vectors3 V23, M23, C23 */

	for (j = 0; j < 4; j++) {
	  V23[ARCHvertex[i][j]][0] += v[j * 3];
	  V23[ARCHvertex[i][j]][1] += v[j * 3 + 1];
	  V23[ARCHvertex[i][j]][2] += v[j * 3 + 2];
	  M23[ARCHvertex[i][j]][0] += Mexv[j * 3];
	  M23[ARCHvertex[i][j]][1] += Mexv[j * 3 + 1];
	  M23[ARCHvertex[i][j]][2] += Mexv[j * 3 + 2];
	  C23[ARCHvertex[i][j]][0] += Cexv[j * 3];
	  C23[ARCHvertex[i][j]][1] += Cexv[j * 3 + 1];
	  C23[ARCHvertex[i][j]][2] += Cexv[j * 3 + 2];
	}

      }

      /* Assemble vectors3 Me, Ce and matrix3 Ke */

      for (j = 0; j < 4; j++) {
	M[ARCHvertex[i][j]][0] += Me[j * 3];
	M[ARCHvertex[i][j]][1] += Me[j * 3 + 1];
	M[ARCHvertex[i][j]][2] += Me[j * 3 + 2];
	C[ARCHvertex[i][j]][0] += Ce[j * 3];
	C[ARCHvertex[i][j]][1] += Ce[j * 3 + 1];
	C[ARCHvertex[i][j]][2] += Ce[j * 3 + 2];
	for (k = 0; k < 4; k++) {
	  if (ARCHvertex[i][j] <= ARCHvertex[i][k]) {
	    kk = ARCHmatrixindex[ARCHvertex[i][j]];
	    while (ARCHmatrixcol[kk] != ARCHvertex[i][k]) {
	      kk++;
	      if (kk >= ARCHmatrixindex[ARCHvertex[i][k] + 1]) {
		printf("K indexing error!!! %d %d\n", ARCHvertex[i][j], 
		       ARCHvertex[i][k]);
		exit(1);
	      }
	    }
	    for (ii = 0; ii < 3; ii++)
	      for (jj = 0; jj < 3; jj++)
		K[kk][ii*3+jj] += Ke[j * 3 + ii][k * 3 + jj];
	  }
	}
      }
    }
  }
  /* end of array initialization by process 0 */

  /* divide arrays over all processes */
  MPI_Bcast( &ARCHnodes, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Exc.dt), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Exc.duration), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Exc.t0), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &timesteps, 1, MPI_INT, 0, MPI_COMM_WORLD );

#ifdef DBGDUMP
  sprintf( fname, "equake%d.dbg", myRank );
  dbgFile = fopen( fname, "w" );
#endif

  distribute_arrays();

  /* exit(0); */


/* Time integration loop */

  /* fprintf(stderr, "\n"); */
  double sum0,sum1,sum2;
  int Anext,Alast,col;
  int stride1 = 3,stride2 = 9;
  int j_1,j_2,i_1,i_2,i_3,j_3,i_4,j_4,i_5,j_5,i_6,j_6,i_7,j_7,i_8,j_8,i_9,j_9,i_10,j_10,i_11,j_11,i_12,j_12,i_13,j_13,i_14,j_14,i_15,j_15,j_dbg;
  int Anext_2,col_2,Anext_7,col_7,Anext_12,col_12;
  int nextPrintedTimestep = 0;

  // timesteps = 3125;
  /* printf("%d timesteps\n", timesteps); */
  double orig_time_start = rtclock();
  for (iter = 1; iter <= timesteps; ) {

    /* if ( (myRank == 0) && (iter > nextPrintedTimestep) ) { */
    /*   printf("timestep %d\n", iter); */
    /*   fflush(stderr); */
    /*   nextPrintedTimestep += 3; */
    /* } */

    for (i_1 = 0; i_1 < ARCHnodes+totalGhosts; i_1++)
      for (j_1 = 0; j_1 < 3; j_1++)
        disp0[i_1][j_1] = 0.0;

    /* loop 2 has an indirect read from disp1, so update disp1 ghosts */
    // printf("update disp1 ghosts\n");
    //fflush(stderr);
    update_disp_ghosts(1);

    for (i_2 = 0; i_2 < ARCHnodes; i_2++) {
    
      Anext_2 = ARCHmatrixindex[i_2]; // start of row in CSR of adjacency matrix for this node

      sum0 = K[Anext_2][0]*disp1[i_2][0] + K[Anext_2][1]*disp1[i_2][1] + K[Anext_2][2]*disp1[i_2][2];
      sum1 = K[Anext_2][3]*disp1[i_2][0] + K[Anext_2][4]*disp1[i_2][1] + K[Anext_2][5]*disp1[i_2][2];
      sum2 = K[Anext_2][6]*disp1[i_2][0] + K[Anext_2][7]*disp1[i_2][1] + K[Anext_2][8]*disp1[i_2][2];

      for( j_2 = ARCHmatrixindex[i_2]+1 ; j_2 < ARCHmatrixindex[i_2+1] ; j_2++){
    	col_2 = ARCHmatrixcol[j_2];

    	sum0 += K[j_2][0]*disp1[col_2][0] + K[j_2][1]*disp1[col_2][1] + K[j_2][2]*disp1[col_2][2];
    	sum1 += K[j_2][3]*disp1[col_2][0] + K[j_2][4]*disp1[col_2][1] + K[j_2][5]*disp1[col_2][2];
    	sum2 += K[j_2][6]*disp1[col_2][0] + K[j_2][7]*disp1[col_2][1] + K[j_2][8]*disp1[col_2][2];
      

    	disp0[col_2][0] += K[j_2][0]*disp1[i_2][0] + K[j_2][3]*disp1[i_2][1] + K[j_2][6]*disp1[i_2][2];
    	disp0[col_2][1] += K[j_2][1]*disp1[i_2][0] + K[j_2][4]*disp1[i_2][1] + K[j_2][7]*disp1[i_2][2];
    	disp0[col_2][2] += K[j_2][2]*disp1[i_2][0] + K[j_2][5]*disp1[i_2][1] + K[j_2][8]*disp1[i_2][2];
      }
      disp0[i_2][0] += sum0;
      disp0[i_2][1] += sum1;
      disp0[i_2][2] += sum2;
    }

    MPI_Barrier(MPI_COMM_WORLD);

    /* add ghost values to disp0 */
    //printf("add disp0 ghosts\n");
    //fflush(stderr);
    add_ghosts_to_disp(0);

    MPI_Barrier(MPI_COMM_WORLD);

    time = iter * Exc.dt;


    for (i_3 = 0; i_3 < ARCHnodes; i_3++)
      for (j_3 = 0; j_3 < 3; j_3++)
        disp0[i_3][j_3] = (-disp0[i_3][j_3] * Exc.dt * Exc.dt ) +  2.0 * M[i_3][j_3] * disp1[i_3][j_3] -
    	  (M[i_3][j_3] - Exc.dt / 2.0 * C[i_3][j_3]) * disp2[i_3][j_3] -
    	    Exc.dt * Exc.dt * (M23[i_3][j_3] * phi2(time) / 2.0 +
    			       C23[i_3][j_3] * phi1(time) / 2.0 +
    			       V23[i_3][j_3] * phi0(time) / 2.0);


    for (i_4 = 0; i_4 < ARCHnodes; i_4++)
      for (j_4 = 0; j_4 < 3; j_4++)
        disp0[i_4][j_4] = disp0[i_4][j_4] /
    	                        (M[i_4][j_4] + Exc.dt / 2.0 * C[i_4][j_4]);

    for (i_5 = 0; i_5 < ARCHnodes; i_5++)
      for (j_5 = 0; j_5 < 3; j_5++)
        vel[i_5][j_5] = 0.5 / Exc.dt * (disp0[i_5][j_5] -
    				    disp2[i_5][j_5]);

    iter += 1;
    /* Print out the response at the source and epicenter nodes */


    for (i_6 = 0; i_6 < ARCHnodes+totalGhosts; i_6++)
      for (j_6 = 0; j_6 < 3; j_6++)
        disp2[i_6][j_6] = 0.0;

     /* loop 7 has an indirect read from disp0, so update disp1 ghosts */
    update_disp_ghosts(0);

    MPI_Barrier(MPI_COMM_WORLD);

    for (i_7 = 0; i_7 < ARCHnodes; i_7++) {
      Anext_7 = ARCHmatrixindex[i_7];

      sum0 = K[Anext_7][0]*disp0[i_7][0] + K[Anext_7][1]*disp0[i_7][1] + K[Anext_7][2]*disp0[i_7][2];
      sum1 = K[Anext_7][3]*disp0[i_7][0] + K[Anext_7][4]*disp0[i_7][1] + K[Anext_7][5]*disp0[i_7][2];
      sum2 = K[Anext_7][6]*disp0[i_7][0] + K[Anext_7][7]*disp0[i_7][1] + K[Anext_7][8]*disp0[i_7][2];

      for( j_7 = ARCHmatrixindex[i_7]+1 ; j_7 < ARCHmatrixindex[i_7+1] ; j_7++){
    	col_7 = ARCHmatrixcol[j_7];

    	sum0 += K[j_7][0]*disp0[col_7][0] + K[j_7][1]*disp0[col_7][1] + K[j_7][2]*disp0[col_7][2];
    	sum1 += K[j_7][3]*disp0[col_7][0] + K[j_7][4]*disp0[col_7][1] + K[j_7][5]*disp0[col_7][2];
    	sum2 += K[j_7][6]*disp0[col_7][0] + K[j_7][7]*disp0[col_7][1] + K[j_7][8]*disp0[col_7][2];
      

    	disp2[col_7][0] += K[j_7][0]*disp0[i_7][0] + K[j_7][3]*disp0[i_7][1] + K[j_7][6]*disp0[i_7][2];
    	disp2[col_7][1] += K[j_7][1]*disp0[i_7][0] + K[j_7][4]*disp0[i_7][1] + K[j_7][7]*disp0[i_7][2];
    	disp2[col_7][2] += K[j_7][2]*disp0[i_7][0] + K[j_7][5]*disp0[i_7][1] + K[j_7][8]*disp0[i_7][2];
      }
      disp2[i_7][0] += sum0;
      disp2[i_7][1] += sum1;
      disp2[i_7][2] += sum2;
    }

    /* add ghost values to disp0 */
    MPI_Barrier(MPI_COMM_WORLD);
    add_ghosts_to_disp(2);
    MPI_Barrier(MPI_COMM_WORLD);


    time = iter * Exc.dt;

    for (i_8 = 0; i_8 < ARCHnodes; i_8++)
      for (j_8 = 0; j_8 < 3; j_8++)
        disp2[i_8][j_8] = (-disp2[i_8][j_8] * Exc.dt * Exc.dt ) +  2.0 * M[i_8][j_8] * disp0[i_8][j_8] -
    	  (M[i_8][j_8] - Exc.dt / 2.0 * C[i_8][j_8]) * disp1[i_8][j_8] -
    	    Exc.dt * Exc.dt * (M23[i_8][j_8] * phi2(time) / 2.0 +
    			       C23[i_8][j_8] * phi1(time) / 2.0 +
    			       V23[i_8][j_8] * phi0(time) / 2.0);


    for (i_9 = 0; i_9 < ARCHnodes; i_9++)
      for (j_9 = 0; j_9 < 3; j_9++)
        disp2[i_9][j_9] = disp2[i_9][j_9] /
    	                        (M[i_9][j_9] + Exc.dt / 2.0 * C[i_9][j_9]);


    for (i_10 = 0; i_10 < ARCHnodes; i_10++)
      for (j_10 = 0; j_10 < 3; j_10++)
        vel[i_10][j_10] = 0.5 / Exc.dt * (disp2[i_10][j_10] -
    				    disp1[i_10][j_10]);

    iter += 1;



    for (i_11 = 0; i_11 < ARCHnodes+totalGhosts; i_11++)
      for (j_11 = 0; j_11 < 3; j_11++)
        disp1[i_11][j_11] = 0.0;

    /* loop 12 has an indirect read from disp2, so update disp2 ghosts */
    MPI_Barrier(MPI_COMM_WORLD);
    update_disp_ghosts(2);
    MPI_Barrier(MPI_COMM_WORLD);



    for (i_12 = 0; i_12 < ARCHnodes; i_12++) {
      Anext_12 = ARCHmatrixindex[i_12];

      sum0 = K[Anext_12][0]*disp2[i_12][0] + K[Anext_12][1]*disp2[i_12][1] + K[Anext_12][2]*disp2[i_12][2];
      sum1 = K[Anext_12][3]*disp2[i_12][0] + K[Anext_12][4]*disp2[i_12][1] + K[Anext_12][5]*disp2[i_12][2];
      sum2 = K[Anext_12][6]*disp2[i_12][0] + K[Anext_12][7]*disp2[i_12][1] + K[Anext_12][8]*disp2[i_12][2];

      for( j_12 = ARCHmatrixindex[i_12]+1 ; j_12 < ARCHmatrixindex[i_12+1] ; j_12++){
    	col_12 = ARCHmatrixcol[j_12];

    	sum0 += K[j_12][0]*disp2[col_12][0] + K[j_12][1]*disp2[col_12][1] + K[j_12][2]*disp2[col_12][2];
    	sum1 += K[j_12][3]*disp2[col_12][0] + K[j_12][4]*disp2[col_12][1] + K[j_12][5]*disp2[col_12][2];
    	sum2 += K[j_12][6]*disp2[col_12][0] + K[j_12][7]*disp2[col_12][1] + K[j_12][8]*disp2[col_12][2];
      

    	disp1[col_12][0] += K[j_12][0]*disp2[i_12][0] + K[j_12][3]*disp2[i_12][1] + K[j_12][6]*disp2[i_12][2];
    	disp1[col_12][1] += K[j_12][1]*disp2[i_12][0] + K[j_12][4]*disp2[i_12][1] + K[j_12][7]*disp2[i_12][2];
    	disp1[col_12][2] += K[j_12][2]*disp2[i_12][0] + K[j_12][5]*disp2[i_12][1] + K[j_12][8]*disp2[i_12][2];
      }
      disp1[i_12][0] += sum0;
      disp1[i_12][1] += sum1;
      disp1[i_12][2] += sum2;
    }

    /* add ghost values to disp1 */
    MPI_Barrier(MPI_COMM_WORLD);
    add_ghosts_to_disp(1);
    MPI_Barrier(MPI_COMM_WORLD);

    time = iter * Exc.dt;


    for (i_13 = 0; i_13 < ARCHnodes; i_13++)
      for (j_13 = 0; j_13 < 3; j_13++)
        disp1[i_13][j_13] = (-disp1[i_13][j_13] * Exc.dt * Exc.dt ) +  2.0 * M[i_13][j_13] * disp2[i_13][j_13] -
    	  (M[i_13][j_13] - Exc.dt / 2.0 * C[i_13][j_13]) * disp0[i_13][j_13] -
    	    Exc.dt * Exc.dt * (M23[i_13][j_13] * phi2(time) / 2.0 +
    			       C23[i_13][j_13] * phi1(time) / 2.0 +
    			       V23[i_13][j_13] * phi0(time) / 2.0);


    for (i_14 = 0; i_14 < ARCHnodes; i_14++)
      for (j_14 = 0; j_14 < 3; j_14++)
        disp1[i_14][j_14] = disp1[i_14][j_14] /
    	                        (M[i_14][j_14] + Exc.dt / 2.0 * C[i_14][j_14]);


    for (i_15 = 0; i_15 < ARCHnodes; i_15++)
      for (j_15 = 0; j_15 < 3; j_15++)
        vel[i_15][j_15] = 0.5 / Exc.dt * (disp1[i_15][j_15] -
    				    disp0[i_15][j_15]);

    iter += 1;


    /* if (iter % Step_stride == 0) { */

/*       fprintf(stderr, "Time step %d\n", iter); */

/*       if (Src.sourcenode <= ARCHmine) */
/* 	printf("%d: %.2e %.2e %.2e\n", ARCHglobalnode[Src.sourcenode], */
/* 	       disp0[Src.sourcenode][0],  */
/* 	       disp0[Src.sourcenode][1], */
/* 	       disp0[Src.sourcenode][2]); */

/*       if (Src.epicenternode <= ARCHmine) */
/* 	printf("%d: %.2e %.2e %.2e\n", ARCHglobalnode[Src.epicenternode],  */
/* 	       disp0[Src.epicenternode][0], */
/* 	       disp0[Src.epicenternode][1], */
/* 	       disp0[Src.epicenternode][2]); */

/*       fflush(stdout); */
    /* } */

/*     i = disptminus; */
/*     disptminus = dispt; */
/*     dispt = disptplus; */
/*     disptplus = i; */

  }

#ifdef DBGDUMP
  fclose( dbgFile );
#endif

  if ( myRank == 0 ) {
    double orig_time_stop = rtclock();
    printf("MX:Manual:%lf\n",orig_time_stop - orig_time_start);
    fflush(stdout);
  }
  MPI_Barrier(MPI_COMM_WORLD);

  fprintf(stderr, "Time step %d\n", iter);

  if ( myRank == sourcenodeRank )
    printf("%d: %.2e %.2e %.2e \n", Src.sourcenode,
	   disp0[Src.sourcenode][0], 
	   disp0[Src.sourcenode][1],
	   disp0[Src.sourcenode][2]);

  if ( myRank == epicenternodeRank ) {
    printf("epicenternode in proc %d, index %d\n", myRank, Src.epicenternode);
    fflush(stderr);
    printf("%d: %.2e %.2e %.2e \n", Src.epicenternode,
	   disp0[Src.epicenternode][0],
	   disp0[Src.epicenternode][1],
	   disp0[Src.epicenternode][2]);
  }
  
    /* FILE* outfile = fopen("orig_output.dat","w"); */
    /* int j_bit; */
    /* char* source_node = (char*)disp0[Src.sourcenode]; */
    /* int in_binary[64]; */
    /* fprintf(outfile,"Source :"); */
    /* int bin_count = 0; */
    /* for( i = 0 ; i< 3*sizeof(double) ; i++){ */
    /*   j_bit = 1; */
    /*   for( j = 0 ; j < 8 ; j++ ){ */
    /*     in_binary[63 - bin_count++] = (source_node[i] & j_bit ? 1 : 0); */
    /*     j_bit = j_bit << 1; */
    /*   } */
    /*   if(bin_count == 64 ){ */
    /*     for( j = 0 ; j < bin_count ; j++ ) */
    /* 	fprintf(outfile,"%d",in_binary[j]); */
    /*     fprintf(outfile," "); */
    /*     bin_count = 0; */
    /*   } */
    /* } */
    
    /* //fprintf(outfile," %x",source_node[i]); */
    /* char* epicenter_node = (char*)disp0[Src.epicenternode]; */
    /* fprintf(outfile,"\nEpicenter :"); */
    /* for( i = 0 ; i< 3*sizeof(double) ; i++){ */
    /*   j_bit = 1; */
    /*   for( j = 0 ; j < 8 ; j++ ){ */
    /*     in_binary[63 - bin_count++] = (source_node[i] & j_bit ? 1 : 0); */
    /*     j_bit = j_bit << 1; */
    /*   } */
    /*   if(bin_count == 64 ){ */
    /*     for( j = 0 ; j < bin_count ; j++ ) */
    /* 	fprintf(outfile,"%d",in_binary[j]); */
    /*     fprintf(outfile," "); */
    /*     bin_count = 0; */
    /*   } */
    /* } */
    /* fprintf(outfile,"\n"); */
    /* fclose(outfile); */

  if ( myRank == 0 ) {
    fprintf(stderr, "%s: %d nodes %d elems %d timesteps\n", 
	    progname, ARCHglobalnodes, ARCHglobalelems, timesteps);
    fprintf(stderr, "\n");
    fflush(stderr);
    
    if (!options.quiet) {
      fprintf(stderr, "%s: Done. Terminating the simulation.\n", progname);
    }
  }

  MPI_Finalize();

  return 0;
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Compute shape function derivatives                                        */
/* N_1 = 1 - r - s - t                                                       */
/* N_2 = r                                                                   */
/* N_3 = s                                                                   */
/* N_4 = t                                                                   */

void shape_ders(ds)
double ds[][4];
{
  ds[0][0] = - 1;
  ds[1][0] = - 1;
  ds[2][0] = - 1;
  ds[0][1] = 1;
  ds[1][1] = 0;
  ds[2][1] = 0;
  ds[0][2] = 0;
  ds[1][2] = 1;
  ds[2][2] = 0;
  ds[0][3] = 0;
  ds[1][3] = 0;
  ds[2][3] = 1;
}

/* --------------------------------------------------------------------------*/
/* Calculate Young's modulus E and Poisson's ratio nu,                       */
/* given a pair of compressional (cp) and shear (cs) wave velocities         */

void get_Enu(prop, E, nu)
struct properties *prop;
double *E, *nu;
{
  double ratio;
  
  ratio = prop->cp / prop->cs;
  ratio = ratio * ratio;
  *nu = 0.5 * (ratio - 2.0) / (ratio - 1.0);
  *E = 2.0 * prop->den * prop->cs * prop->cs * (1.0 + *nu);
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the inverse and the determinant of the Jacobian,                */
/* given the Jacobian                                                        */
/* (a on input holds the Jacobian and on output its inverse                  */

void inv_J(a, det)
double a[][3];
double *det;
{
  double d1;
  double c[3][3];
  int i, j;
  
  c[0][0] = a[1][1] * a[2][2] - a[2][1] * a[1][2];
  c[0][1] = a[0][2] * a[2][1] - a[0][1] * a[2][2];
  c[0][2] = a[0][1] * a[1][2] - a[0][2] * a[1][1];
  c[1][0] = a[1][2] * a[2][0] - a[1][0] * a[2][2];
  c[1][1] = a[0][0] * a[2][2] - a[0][2] * a[2][0];
  c[1][2] = a[0][2] * a[1][0] - a[0][0] * a[1][2];
  c[2][0] = a[1][0] * a[2][1] - a[1][1] * a[2][0];
  c[2][1] = a[0][1] * a[2][0] - a[0][0] * a[2][1];
  c[2][2] = a[0][0] * a[1][1] - a[0][1] * a[1][0];
  *det = a[0][0] * c[0][0] + a[0][1] * c[1][0] + a[0][2] * c[2][0];
  d1 = 1.0 / *det;
  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      a[i][j] = c[i][j] * d1;
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the element stiffness (Ke[12][12]) and                          */
/*           the element mass matrices (Me[12]),                             */
/* given the four vertices of a tetrahedron                                  */

void element_matrices(vertices, prop, Ke, Me)
double vertices[][3], Ke[][12], Me[];
struct properties *prop;
{
  double ds[3][4];
  double sum[3];
  double jacobian[3][3];
  double det;
  double volume;
  double E, nu;
  double c1, c2, c3;
  double tt, ts;
  int i, j, m, n, row, column;
  
  shape_ders(ds);

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++) {
      sum[0] = 0.0;
      for (m = 0; m < 4; m++)
        sum[0] = sum[0] + ds[i][m] * vertices[m][j];
      jacobian[j][i] = sum[0];         /* compute Jacobian */
    }

  inv_J(jacobian, &det);               /* compute J^-1 & its determinant */

  for (m = 0; m < 4; m++) {

    for (i = 0; i < 3; i++) {
      sum[i] = 0.0;
      for (j = 0; j < 3; j++)
	sum[i] = sum[i] + jacobian[j][i] * ds[j][m];
    }

    for (i = 0; i < 3; i++) 
      ds[i][m] = sum[i];

  }

  volume = det / 6.0;

  if (volume <= 0) {
    fprintf(stderr, "Warning: Element volume = %f !\n", volume);
  }

  get_Enu(prop, &E, &nu);

  c1 = E / (2.0 * (nu + 1.0) * (1.0 - nu * 2.0)) * volume;
  c2 = E * nu / ((nu + 1.0) * (1.0 - nu * 2.0)) * volume;
  c3 = E / ((nu + 1.0) * 2.0) * volume;

  row = - 1;

  for (m = 0; m < 4; m++) {            /* lower triangular stiffness matrix */
    for (i = 0; i < 3; ++i) {
      ++row;
      column = - 1;
      for (n = 0; n <= m; n++) {
        for (j = 0; j < 3; j++) {
          ++column;
          ts = ds[i][m] * ds[j][n];
          if (i == j) {
            ts = ts * c1;
            tt = (ds[0][m] * ds[0][n] +
		  ds[1][m] * ds[1][n] +
		  ds[2][m] * ds[2][n]) * c3;
          }
	  else {
            if (m == n) {
              ts = ts * c1;
              tt = 0;
            }
	    else {
              ts = ts * c2;
              tt = ds[j][m] * ds[i][n] * c3;
            }
          }
          Ke[row][column] = Ke[row][column] + ts + tt;
        }
      }
    }
  }
  tt = prop->den * volume / 4.0;
  for (i = 0; i < 12; i++) 
    Me[i] = tt;

  for (i = 0; i < 12; i++)
    for (j = 0; j <= i; j++) 
      Ke[j][i] = Ke[i][j];
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the area of a triangle given the coordinates of                 */
/* its three vertices                                                        */

double area_triangle(vertices)
double vertices[][3];
{
  double a, b, c;
  double x2, y2, z2;
  double p;
  double area;
  
  x2 = (vertices[0][0] - vertices[1][0]) * (vertices[0][0] - vertices[1][0]);
  y2 = (vertices[0][1] - vertices[1][1]) * (vertices[0][1] - vertices[1][1]);
  z2 = (vertices[0][2] - vertices[1][2]) * (vertices[0][2] - vertices[1][2]);
  a = sqrt(x2 + y2 + z2);

  x2 = (vertices[2][0] - vertices[1][0]) * (vertices[2][0] - vertices[1][0]);
  y2 = (vertices[2][1] - vertices[1][1]) * (vertices[2][1] - vertices[1][1]);
  z2 = (vertices[2][2] - vertices[1][2]) * (vertices[2][2] - vertices[1][2]);
  b = sqrt(x2 + y2 + z2);

  x2 = (vertices[0][0] - vertices[2][0]) * (vertices[0][0] - vertices[2][0]);
  y2 = (vertices[0][1] - vertices[2][1]) * (vertices[0][1] - vertices[2][1]);
  z2 = (vertices[0][2] - vertices[2][2]) * (vertices[0][2] - vertices[2][2]);
  c = sqrt(x2 + y2 + z2);

  p = (a + b + c) / 2.0;

  area = sqrt(p * (p - a) * (p - b) * (p - c));

  return area;

}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Generate the element damping matrix for the absorbing boundary            */
/* (triangular plane element)                                                */

void abe_matrix(vertices, bv, prop, Ce)
double vertices[][3];
int bv[];
struct properties *prop;
double Ce[];
{
  int i, j;
  double area;
  
  area = area_triangle(vertices);

  for (i = 0; i < 3; i++) {
    j = 3 * bv[i];
    Ce[j] = Ce[j] + prop->cs * prop->den * area / 3.0;
    Ce[j + 1] = Ce[j + 1] + prop->cs * prop->den * area / 3.0;
    Ce[j + 2] = Ce[j + 2] + prop->cp * prop->den * area / 3.0;
  }

}
/* --------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------*/
/* Excitation (ramp function)                                                */

double phi0(t)
double t;
{
  double value;
  
  if (t <= Exc.t0) {

    value = 0.5 / PI * (2.0 * PI * t / Exc.t0 - sin(2.0 * PI * t / Exc.t0));
    return value;

  } 
  else
    return 1.0;

}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* First derivative of the excitation (velocity of ramp function)            */

double phi1(t)
double t;
{
  double value;
  
  if (t <= Exc.t0) {

    value = (1.0 - cos(2.0 * PI * t / Exc.t0)) / Exc.t0;
    return value;

  }
  else
    return 0.0;
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Second derivative of the excitation (acceleration of ramp function)       */

double phi2(t)
double t;
{
  double value;
  
  if (t <= Exc.t0) {

    value = 2.0 * PI / Exc.t0 / Exc.t0 * sin(2.0 * PI * t / Exc.t0);
    return value;

  }
  else
    return 0.0;
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the slip motion at the source node                              */

void slip(u, v, w)
double *u, *v, *w;
{
  *u = *v = *w = 0.0;
  *u = (cos(Src.rake) * sin(Src.strike) - 
	sin(Src.rake) * cos(Src.strike) * cos(Src.dip));
  *v = (cos(Src.rake) * cos(Src.strike) +
	sin(Src.rake) * sin(Src.strike) * cos(Src.dip));
  *w = sin(Src.rake) * sin(Src.dip);
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the distance between two points p1 and p2                       */

double distance(p1, p2)
double p1[], p2[];
{
  return ((p1[0] - p2[0]) * (p1[0] - p2[0]) +
	  (p1[1] - p2[1]) * (p1[1] - p2[1]) +
	  (p1[2] - p2[2]) * (p1[2] - p2[2]));
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the centroid of a tetrahedron                                   */

void centroid(x, xc)
double x[][3], xc[];
{

  int i;
  
  for (i = 0; i < 3; i++)
    xc[i] = (x[0][i] + x[1][i] + x[2][i] + x[3][i]) / 4.0;

}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Calculate the distance to the fault from a given point x                 */

double point2fault(x)
double x[];
{

  double nx, ny, nz;
  double d0;
  
  nx = cos(Src.strike) * sin(Src.dip);
  ny = - sin(Src.strike) * sin(Src.dip);
  nz = cos(Src.dip);

  d0 = - (nx * Src.xyz[0] + ny * Src.xyz[1] + nz * Src.xyz[2]);

  return (double) nx * x[0] + ny * x[1] + nz * x[2] + d0;
}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Matrix (12x12) times vector (12x1) product                                */

void mv12x12(m, v)
double m[][12], v[];
{
  int i, j;
  double u[12];
  
  for (i = 0; i < 12; i++) {

    u[i] = 0;
    for (j = 0; j < 12; j++) 
      u[i] += m[i][j] * v[j];
  }

  for (i = 0; i < 12; i++) 
    v[i] = u[i];

}
/* --------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------*/
/* Vector (12x1) times vector (12x1) product                                 */

void vv12x12(v1, v2, u)
double v1[], v2[], u[];
{
  int i;
  
  for (i = 0; i < 12; i++) 
    u[i] = v1[i] * v2[i];

}

/*--------------------------------------------------------------------------*/ 
/* Graceful exit                                                            */

void arch_bail(void) {
    exit(0);
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
void arch_info(void)
{
    printf("\n");
    printf("You are running an Archimedes finite element simulation called %s.\n\n", progname);
    printf("The command syntax is:\n\n");
    printf("%s [-Qh] < packfile\n\n", progname);
    printf("Command line options:\n\n");
    printf("    -Q  Quietly suppress all explanation of what this program is doing\n");
    printf("        unless an error occurs.\n");
    printf("    -h  Print this message and exit.\n");
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
/*
 * arch_parsecommandline - parse the command line
 */
void arch_parsecommandline(int argc, char **argv, struct options *op) 
{
    int i, j;
    
    /* first set up the defaults */
    op->quiet = 0;
    op->help = 0;

    /* now see if the user wants to change any of these */
    for (i=1; i<argc; i++) {
	if (argv[i][0] == '-') {
	    for (j = 1; argv[i][j] != '\0'; j++) {
		if (argv[i][j] == 'Q') {
		    op->quiet = 1;
		}
		if ((argv[i][j] == 'h' ||argv[i][j] == 'H')) {
		    op->help = 1;
		}
	    }
	}
    }
    if (op->help) {
	arch_info();
	exit(0);
    }
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
/*
 * arch_readnodevector - read a vector of nodal data from the pack file
 *                       called by READNODEVECTOR.stub               
 */
void arch_readnodevector(double *v, int n) {
    int i;
    int type, attributes;
    
    fscanf(packfile, "%d %d\n", &type, &attributes);

    if (type != 2) {
	fprintf(stderr, 
		"READNODEVECTOR: unexpected data type\n");
	arch_bail();
    }
    if (attributes != 1) {
	fprintf(stderr, 
		"READNODEVECTOR: unexpected number of attributes\n");
	arch_bail();
    }
    for (i=0; i<n; i++) {	
	fscanf(packfile, "%lf", &v[i]);
    }
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
/*
 * arch_readelemvector - read a vector of element data from the pack file
 *                       called by READELEMVECTOR.stub               
 */
void arch_readelemvector(double *v, int n) {
    int i;
    int type, attributes;
    
    fscanf(packfile, "%d %d\n", &type, &attributes);
    if (type != 1) {
	fprintf(stderr, 
		"READELEMVECTOR: unexpected data type\n");
	arch_bail();
    }
    if (attributes != 1) {
	fprintf(stderr, 
		"READELEMVECTOR: unexpected number of attributes\n");
	arch_bail();
    }
    for (i=0; i<n; i++) {	
	fscanf(packfile, "%lf", &v[i]);
    }
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
/*
 * arch_readdouble - read a floating point number from the pack file
 */
void arch_readdouble(double *v) {
    int type, attributes;
    
    fscanf(packfile, "%d %d\n", &type, &attributes);
    if (type != 3) {
	fprintf(stderr, 
		"READDOUBLE: unexpected data type\n");
	arch_bail();
    }
    if (attributes != 1) {
	fprintf(stderr, 
		"READDOUBLE: unexpected number of attributes\n");
	arch_bail();
    }
    fscanf(packfile, "%lf", &v[0]);
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
void readpackfile(FILE *packfile, struct options *op) {
  int oldrow, newrow;
  int i, j;
  int temp1, temp2;

  fscanf(packfile, "%d", &ARCHglobalnodes);
  fscanf(packfile, "%d", &ARCHmesh_dim);
  fscanf(packfile, "%d", &ARCHglobalelems);
  fscanf(packfile, "%d", &ARCHcorners);
  fscanf(packfile, "%d", &ARCHsubdomains);
  fscanf(packfile, "%lf", &ARCHduration);

  /* only one subdomain allowed */
  if (ARCHsubdomains != 1) {
    fprintf(stderr, "%s: too many subdomains(%d), rerun slice using -s1\n",
	    progname, ARCHsubdomains);
    arch_bail();
  }

  /* read nodes */
  if (!op->quiet) {
    fprintf(stderr, "%s: Reading nodes.\n", progname);
  }

  fscanf(packfile, "%d %d %d", &ARCHnodes, &ARCHmine, &ARCHpriv);

  ARCHglobalnode = (int *) malloc(ARCHnodes * sizeof(int));
  if (ARCHglobalnode == (int *) NULL) {
    fprintf(stderr, "malloc failed for ARCHglobalnode\n");
    fflush(stderr);
    exit(0);
  }

  /* ARCHcoord = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* for (i = 0; i < ARCHnodes; i++) */
  /*   ARCHcoord[i] = (double *) malloc(3 * sizeof(double)); */
  ARCHcoord = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));

  for (i=0; i<ARCHnodes; i++) {
    fscanf(packfile, "%d", &ARCHglobalnode[i]);
    for (j=0; j<ARCHmesh_dim; j++) {
      fscanf(packfile, "%lf", &ARCHcoord[i][j]);
    }
  }

  /* read elements */
  if (!op->quiet)
    fprintf(stderr, "%s: Reading elements.\n", progname);

  fscanf(packfile, "%d", &ARCHelems);

  ARCHglobalelem = (int *) malloc(ARCHelems * sizeof(int));
  if (ARCHglobalelem == (int *) NULL) {
    fprintf(stderr, "malloc failed for ARCHglobalelem\n");
    fflush(stderr);
    exit(0);
  }

  /* ARCHvertex = (int **) malloc(ARCHelems * sizeof(int *)); */
  /* for (i = 0; i < ARCHelems; i++) */
  /*   ARCHvertex[i] = (int *) malloc(4 * sizeof(int)); */
  ARCHvertex = (int (*)[4])malloc(sizeof(int)*4*ARCHelems);

  for (i=0; i<ARCHelems; i++) {
    fscanf(packfile, "%d", &ARCHglobalelem[i]);	
    for (j=0; j<ARCHcorners; j++) {
      fscanf(packfile, "%d", &ARCHvertex[i][j]);
    }
  }
    
  /* read sparse matrix structure and convert from tuples to CSR */
  if (!op->quiet)
    fprintf(stderr, "%s: Reading sparse matrix structure.\n", progname);

  fscanf(packfile, "%d %d", &ARCHmatrixlen, &ARCHcholeskylen);

  ARCHmatrixcol = (int *) malloc((ARCHmatrixlen + 1) * sizeof(int));
  if (ARCHmatrixcol == (int *) NULL) {
    fprintf(stderr, "malloc failed for ARCHmatrixcol\n");
    fflush(stderr);
    exit(0);
  }

  ARCHmatrixindex = (int *) malloc((ARCHnodes + 1) * sizeof(int));
  if (ARCHmatrixindex == (int *) NULL) {
    fprintf(stderr, "malloc failed for ARCHmatrixindex\n");
    fflush(stderr);
    exit(0);
  }

  oldrow = -1; 
  for (i = 0; i < ARCHmatrixlen; i++) {
    fscanf(packfile, "%d", &newrow);
    fscanf(packfile, "%d", &ARCHmatrixcol[i]);
    while (oldrow < newrow) { 
      if (oldrow+1 >= ARCHnodes+1) {
	printf("%s: error: (1)idx buffer too small (%d >= %d)\n", 
	       progname, oldrow+1, ARCHnodes+1);
	arch_bail();
      }
      ARCHmatrixindex[++oldrow] = i;
    }
  }
  while (oldrow < ARCHnodes) {
    ARCHmatrixindex[++oldrow] = ARCHmatrixlen;
  }
    
  /* read comm info (which nodes are shared between subdomains) */
  fscanf(packfile, "%d %d", &temp1, &temp2);
    
}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
/* 
 * arch_init - initialize the Archimedes simulation 
 *             called by ARCHIMEDES_INIT.stub
 */
void arch_init(int argc, char **argv, struct options *op)
{

  /* parse the command line options */
  progname = argv[0];
  arch_parsecommandline(argc, argv, op);

  /* read the pack file */
  printf("Open pack file\n");
  packfile = fopen(argv[1],"r");
  if ( packfile == 0 ) {
    printf("No packfile -- aborting\n");
    exit(1);
  }
  printf("Read pack file\n");
  readpackfile(packfile, op);
  printf("Done reading pack file\n");

}
/*--------------------------------------------------------------------------*/ 


/*--------------------------------------------------------------------------*/ 
/* Matrix vector product - basic version                                    */

void smvp(int nodes, double ***A, int *Acol, 
		int *Aindex, double **v, double **w) {
  int i;
  int Anext, Alast, col;
  double sum0, sum1, sum2;
 
  for (i = 0; i < nodes; i++) {
    Anext = Aindex[i];
    Alast = Aindex[i + 1];

    sum0 = A[Anext][0][0]*v[i][0] + A[Anext][0][1]*v[i][1] + A[Anext][0][2]*v[i][2];
    sum1 = A[Anext][1][0]*v[i][0] + A[Anext][1][1]*v[i][1] + A[Anext][1][2]*v[i][2];
    sum2 = A[Anext][2][0]*v[i][0] + A[Anext][2][1]*v[i][1] + A[Anext][2][2]*v[i][2];

    Anext++;
    while (Anext < Alast) {
      col = Acol[Anext];

      sum0 += A[Anext][0][0]*v[col][0] + A[Anext][0][1]*v[col][1] + A[Anext][0][2]*v[col][2];
      sum1 += A[Anext][1][0]*v[col][0] + A[Anext][1][1]*v[col][1] + A[Anext][1][2]*v[col][2];
      sum2 += A[Anext][2][0]*v[col][0] + A[Anext][2][1]*v[col][1] + A[Anext][2][2]*v[col][2];
      

      w[col][0] += A[Anext][0][0]*v[i][0] + A[Anext][1][0]*v[i][1] + A[Anext][2][0]*v[i][2];
      w[col][1] += A[Anext][0][1]*v[i][0] + A[Anext][1][1]*v[i][1] + A[Anext][2][1]*v[i][2];
      w[col][2] += A[Anext][0][2]*v[i][0] + A[Anext][1][2]*v[i][1] + A[Anext][2][2]*v[i][2];
      Anext++;
    }
    w[i][0] += sum0;
    w[i][1] += sum1;
    w[i][2] += sum2;
  }
}


/*--------------------------------------------------------------------------*/ 
/* Matrix vector product - hand optimized with lots of temporaries          */

void smvp_opt(nodes, A, Acol, Aindex, v, w)
int nodes;
double ***A;
int *Acol;
int *Aindex;
double **v;
double **w;
{
  int i;
  int Anext, Alast;
  double vi0, vi1, vi2, sum0, sum1, sum2, value;
  double vcol0, vcol1, vcol2, wcol0, wcol1, wcol2;
  int col;
  
  for (i = 0; i < nodes; i++) {
    w[i][0] = 0.0;
    w[i][1] = 0.0;
    w[i][2] = 0.0;
  }
  for (i = 0; i < nodes; i++) {
    vi0 = v[i][0];
    vi1 = v[i][1];
    vi2 = v[i][2];
    Anext = Aindex[i];
    Alast = Aindex[i + 1];
    sum0 = w[i][0] +
           A[Anext][0][0] * vi0 + A[Anext][0][1] * vi1 + A[Anext][0][2] * vi2;
    sum1 = w[i][1] +
           A[Anext][1][0] * vi0 + A[Anext][1][1] * vi1 + A[Anext][1][2] * vi2;
    sum2 = w[i][2] +
           A[Anext][2][0] * vi0 + A[Anext][2][1] * vi1 + A[Anext][2][2] * vi2;
    Anext++;
    while (Anext < Alast) {
      col = Acol[Anext];
      vcol0 = v[col][0];
      vcol1 = v[col][1];
      vcol2 = v[col][2];
      value = A[Anext][0][0];
      sum0 += value * vcol0;
      wcol0 = w[col][0] + value * vi0;
      value = A[Anext][0][1];
      sum0 += value * vcol1;
      wcol1 = w[col][1] + value * vi0;
      value = A[Anext][0][2];
      sum0 += value * vcol2;
      wcol2 = w[col][2] + value * vi0;
      value = A[Anext][1][0];
      sum1 += value * vcol0;
      wcol0 += value * vi1;
      value = A[Anext][1][1];
      sum1 += value * vcol1;
      wcol1 += value * vi1;
      value = A[Anext][1][2];
      sum1 += value * vcol2;
      wcol2 += value * vi1;
      value = A[Anext][2][0];
      sum2 += value * vcol0;
      w[col][0] = wcol0 + value * vi2;
      value = A[Anext][2][1];
      sum2 += value * vcol1;
      w[col][1] = wcol1 + value * vi2;
      value = A[Anext][2][2];
      sum2 += value * vcol2;
      w[col][2] = wcol2 + value * vi2;
      Anext++;
    }
    w[i][0] = sum0;
    w[i][1] = sum1;
    w[i][2] = sum2;
  }
}

/*--------------------------------------------------------------------------*/ 
/* Dynamic memory allocations and initializations                           */

void mem_init(void) {

  int i, j, k;

  mem_init_basic();
  mem_init_K(); 
  /* Initializations */

  for (i = 0; i < ARCHelems; i++) {
    source_elms[i] = 1;
  }

}

void mem_init_K(void) {

  int i, j, k;

  /* Stiffness matrix K[ARCHmatrixlen][3][3] */

  /* double* K_temp = (double*)malloc(sizeof(double)*9*ARCHmatrixlen); */
  /* K = (double **) malloc(ARCHmatrixlen * sizeof(double **)); */
  /* if (K == (double **) NULL) { */
  K = (double (*)[9]) malloc(ARCHmatrixlen * 9 * sizeof(double));
  if (K == (double (*)[9]) NULL) {
    fprintf(stderr, "malloc failed for K\n");
    fflush(stderr);
    exit(0);
  }
  /* for (i = 0; i < ARCHmatrixlen; i++) { */
  /*   K[i] = K_temp + i*9; */
  /* } */

  /* for (i = 0; i < ARCHmatrixlen; i++) { */
  /*   for (j = 0; j < 3; j++) { */
  /*     for (k = 0; k < 3; k++) { */
  /*       K[i][j*3+k] = 0.0; */
  /*     } */
  /*   } */
  /* } */
  
}

void mem_init_basic(void) {

  int i, j, k;

  /* Node vector */
  nodekindf = (double *) malloc(ARCHnodes * sizeof(double));
  if (nodekindf == (double *) NULL) {
    fprintf(stderr, "malloc failed for nodekindf\n");
    fflush(stderr);
    exit(0);
  }

/* Node vector */

  nodekind = (int *) malloc(ARCHnodes * sizeof(int));
  if (nodekind == (int *) NULL) {
    fprintf(stderr, "malloc failed for nodekind\n");
    fflush(stderr);
    exit(0);
  }

/* Element vector */

  source_elms = (int *) malloc(ARCHelems * sizeof(int));
  if (source_elms == (int *) NULL) {
    fprintf(stderr, "malloc failed for source_elms\n");
    fflush(stderr);
    exit(0);
  }

/* Velocity array */
  
  /* double* vel_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* vel = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (vel == (double **) NULL) { */
  vel = (double (*)[3])malloc(ARCHnodes * sizeof(double)*3);
  if (vel == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for vel\n");
    fflush(stderr);
    exit(0);
  }
  /* for (i = 0; i < ARCHnodes; i++) { */
  /*   vel[i] = vel_temp + i * 3 ;//(double *) malloc(3 * sizeof(double)); */
  /*   if (vel[i] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for vel[%d]\n",i); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

/* Mass matrix */
  
  /* double* M_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* M = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (M == (double **) NULL) { */
  M = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (M == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for M\n");
    fflush(stderr);
    exit(0);
  }
  /*   fprintf(stderr, "malloc failed for M\n"); */
  /*   fflush(stderr); */
  /*   exit(0); */
  /* } */
  /* for (i = 0; i < ARCHnodes; i++) { */
  /*   M[i] = M_temp + i*3; //(double *) malloc(3 * sizeof(double)); */
  /*   if (M[i] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for M[%d]\n",i); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

/* Damping matrix */
  /* double* C_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* C = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (C == (double **) NULL) { */
  C = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (C == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for C\n");
    fflush(stderr);
    exit(0);
  }
  /* for (i = 0; i < ARCHnodes; i++) { */
  /*   C[i] = C_temp+ i*3;//(double *) malloc(3 * sizeof(double)); */
  /*   if (C[i] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for C[%d]\n",i); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

/* Auxiliary mass matrix */
  /* double* M23_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* M23 = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (M23 == (double **) NULL) { */
  M23 = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (M23 == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for M23\n");
    fflush(stderr);
    exit(0);
  }
  /* for (i = 0; i < ARCHnodes; i++) { */
  /*   M23[i] = M23_temp + i*3 ; //(double *) malloc(3 * sizeof(double)); */
  /*   if (M23[i] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for M23[%d]\n",i); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

/* Auxiliary damping matrix */
  /* double* C23_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* C23 = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (C23 == (double **) NULL) { */
  C23 = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (C23 == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for C23\n");
    fflush(stderr);
    exit(0);
  }
  /* for (i = 0; i < ARCHnodes; i++) { */
  /*   C23[i] = C23_temp + i*3 ; //(double *) malloc(3 * sizeof(double)); */
  /*   if (C23[i] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for C23[%d]\n",i); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

/* Auxiliary vector */
  /* double* V23_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* V23 = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (V23 == (double **) NULL) { */
  V23 = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (V23 == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for V23\n");
    fflush(stderr);
    exit(0);
  }
  /* for (i = 0; i < ARCHnodes; i++) { */
  /*   V23[i] = V23_temp + i*3; //(double *) malloc(3 * sizeof(double)); */
  /*   if (V23[i] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for V23[%d]\n",i); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

  /* Displacement array disp[3][ARCHnodes][3] */
  /* disp = (double ***) malloc(3 * sizeof(double **)); */
  /* if (disp == (double ***) NULL) { */
  /*   fprintf(stderr, "malloc failed for disp\n"); */
  /*   fflush(stderr); */
  /*   exit(0); */
  /* } */
  /* double* disp0_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* disp0 = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (disp0 == (double **) NULL) { */
  disp0 = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (disp0 == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for disp0\n");
    fflush(stderr);
    exit(0);
  }
  /* for (j = 0; j < ARCHnodes; j++) { */
  /*   disp0[j] = disp0_temp + j*3 ;//(double *) malloc(3 * sizeof(double)); */
  /*   if (disp0[j] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for disp0[%d]\n",j); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */
  
  /* double* disp1_temp = (double*) malloc(ARCHnodes * 3 * sizeof(double)); */
  /* disp1 = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (disp1 == (double **) NULL) { */
  disp1 = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (disp1 == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for disp1\n");
    fflush(stderr);
    exit(0);
  }
  /* for (j = 0; j < ARCHnodes; j++) { */
  /*   disp1[j] = disp1_temp+j*3;//(double *) malloc(3 * sizeof(double)); */
  /*   if (disp1[j] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for disp1[%d]\n",j); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

  /* double* disp2_temp = (double*) malloc(sizeof(double) * 3 * ARCHnodes); */
  /* disp2 = (double **) malloc(ARCHnodes * sizeof(double *)); */
  /* if (disp2 == (double **) NULL) { */
  disp2 = (double (*)[3]) malloc(ARCHnodes * 3 * sizeof(double));
  if (disp2 == (double (*)[3]) NULL) {
    fprintf(stderr, "malloc failed for disp2\n");
    fflush(stderr);
    exit(0);
  }
  /* for (j = 0; j < ARCHnodes; j++) { */
  /*   disp2[j] = disp2_temp + j*3 ;//(double *) malloc(3 * sizeof(double)); */
  /*   if (disp2[j] == (double *) NULL) { */
  /*     fprintf(stderr, "malloc failed for disp2[%d]\n",j); */
  /*     fflush(stderr); */
  /*     exit(0); */
  /*   } */
  /* } */

  /* Initializations */

  for (i = 0; i < ARCHnodes; i++) {
    nodekind[i] = 0;
    for (j = 0; j < 3; j++) {
      M[i][j] = 0.0;
      C[i][j] = 0.0;
      M23[i][j] = 0.0;
      C23[i][j] = 0.0;
      V23[i][j] = 0.0;
      disp0[i][j] = 0.0;
      disp1[i][j] = 0.0;
      disp2[i][j] = 0.0;
    }
  }

}

void distribute_arrays()
{
  int rank;
  int j, k, index;
  MPI_Status status;
  int *ghostIndex;
  int *packedLocalGhostIndices;
  int *packedRemoteGhostIndices;
  int *packedLGIdisplacements;
  int *packedRGIdisplacements;
  int packedLGIsize, packedRGIsize;
  
  /* determine which portions of the arrays are owned by all processes */
  ARCHglobalnodes = ARCHnodes;
  ARCHstart = (int*)malloc(numProcs*sizeof(int));
  ARCHend = (int*)malloc(numProcs*sizeof(int));
  ARCHsize = (int*)malloc(numProcs*sizeof(int));
  for ( rank = 0; rank < numProcs; rank++ ) {
    ARCHstart[rank] = (ARCHglobalnodes*rank) / numProcs;
    ARCHend[rank] = (ARCHglobalnodes*(rank+1)) / numProcs;
    ARCHsize[rank] = ARCHend[rank] - ARCHstart[rank];
  }

  ARCHnodes = ARCHend[myRank] - ARCHstart[myRank];
  
#ifdef DBGDUMP
  fprintf(dbgFile, "Process %d holds array elements from %d to %d\n",
	  myRank, ARCHstart[myRank], ARCHend[myRank]);
  fflush(dbgFile);
#endif
  
  if ( myRank != 0 ) {
    mem_init_basic();
  }
  
  if ( myRank == 0 ) {
    for ( rank = 1; rank < numProcs; rank++ ) {
#ifdef DBGDUMP
      fprintf(dbgFile, "Sending %d elements to process %d\n", ARCHsize[rank], rank );
      fflush(dbgFile);
#endif
      MPI_Send( &(nodekindf[ARCHstart[rank]]), ARCHsize[rank], MPI_DOUBLE, rank, 1, MPI_COMM_WORLD );
      MPI_Send( &(nodekind[ARCHstart[rank]]), ARCHsize[rank], MPI_INT, rank, 2, MPI_COMM_WORLD );
      MPI_Send( &(vel[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 3, MPI_COMM_WORLD );
      MPI_Send( &(M[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 4, MPI_COMM_WORLD );
      MPI_Send( &(C[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 5, MPI_COMM_WORLD );
      MPI_Send( &(M23[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 6, MPI_COMM_WORLD );
      MPI_Send( &(C23[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 7, MPI_COMM_WORLD );
      MPI_Send( &(V23[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 8, MPI_COMM_WORLD );
      MPI_Send( &(disp0[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 9, MPI_COMM_WORLD );
      MPI_Send( &(disp1[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 10, MPI_COMM_WORLD );
      MPI_Send( &(disp2[ARCHstart[rank]][0]) , 3*ARCHsize[rank], MPI_DOUBLE, rank, 11, MPI_COMM_WORLD );
    }
  }
  else {
#ifdef DBGDUMP
    fprintf(dbgFile, "Receiving %d elements from process 0\n", ARCHnodes );
    fflush(dbgFile);
#endif
    MPI_Recv( nodekindf, ARCHnodes, MPI_DOUBLE, 0, 1, MPI_COMM_WORLD, &status );
    MPI_Recv( nodekind, ARCHnodes, MPI_INT, 0, 2, MPI_COMM_WORLD, &status );
    MPI_Recv( &(vel[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 3, MPI_COMM_WORLD, &status );
    MPI_Recv( &(M[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 4, MPI_COMM_WORLD, &status );
    MPI_Recv( &(C[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 5, MPI_COMM_WORLD, &status );
    MPI_Recv( &(M23[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 6, MPI_COMM_WORLD, &status );
    MPI_Recv( &(C23[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 7, MPI_COMM_WORLD, &status );
    MPI_Recv( &(V23[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 8, MPI_COMM_WORLD, &status );
    MPI_Recv( &(disp0[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 9, MPI_COMM_WORLD, &status );
    MPI_Recv( &(disp1[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 10, MPI_COMM_WORLD, &status );
    MPI_Recv( &(disp2[0][0]), 3*ARCHnodes, MPI_DOUBLE, 0, 11, MPI_COMM_WORLD, &status );
  }

  MPI_Barrier(MPI_COMM_WORLD);

#ifdef DBGDUMP
  fprintf(dbgFile, "Done distributing direct-access arrays\n");
  fflush(dbgFile);
#endif

  /*
   *--------------------------------------------------------------------
   * Now distribute the adjacency matrix structures and determine ghosts
   * Set up arrays needed for communicating ghosts
   *--------------------------------------------------------------------
   */

  /*-----------------*/
  /* ARCHmatrixindex */
  /*-----------------*/
  if ( myRank == 0 ) {
    for ( rank = 1; rank < numProcs; rank++ ) {
      MPI_Send( &(ARCHmatrixindex[ARCHstart[rank]]), ARCHsize[rank]+1, MPI_INT, rank, 12, MPI_COMM_WORLD );
    }
  }
  else {
    ARCHmatrixindex = (int *) malloc((ARCHnodes + 1) * sizeof(int));
    if (ARCHmatrixindex == (int *) NULL) {
      fprintf(stderr, "malloc failed for ARCHmatrixindex, proc %d\n", myRank);
      fflush(stderr);
      exit(0);
    }
    MPI_Recv( ARCHmatrixindex, ARCHnodes+1, MPI_INT, 0, 12, MPI_COMM_WORLD, &status );
    /* Adjust entries of ARCHmatrixindex for local portion */
    for ( j = ARCHnodes; j > 0; j-- )
      ARCHmatrixindex[j] -= ARCHmatrixindex[0];
    ARCHmatrixindex[0] = 0;
  }

#ifdef DBGDUMP
  fprintf(dbgFile, "Done distributing ARCHmatrixindex arrays\n");
  fflush(dbgFile);
#endif

  /*---------------*/
  /* ARCHmatrixcol */
  /*---------------*/

  /* NOTE that process 0 still has the whole adjacency matrix */
  if ( myRank == 0 ) {
    for ( rank = 1; rank < numProcs; rank++ ) {
      int colSize;

#ifdef DBGDUMP
      fprintf(dbgFile, "Send proc %d ARCHmatrixcols from index %d to %d\n",
	      rank, ARCHmatrixindex[ARCHstart[rank]],
	      rank < numProcs-1 ? ARCHmatrixindex[ARCHstart[rank+1]] : ARCHmatrixlen );
      fflush(dbgFile);
#endif

      if ( rank < numProcs-1 )
	colSize = ARCHmatrixindex[ARCHstart[rank+1]] - ARCHmatrixindex[ARCHstart[rank]];
      else
	colSize = ARCHmatrixlen - ARCHmatrixindex[ARCHstart[rank]];
      MPI_Send( &colSize, 1, MPI_INT, rank, 13, MPI_COMM_WORLD );
      MPI_Send( &(ARCHmatrixcol[ARCHmatrixindex[ARCHstart[rank]]]), colSize, MPI_INT, rank, 14, MPI_COMM_WORLD );
      MPI_Send( &(K[ARCHmatrixindex[ARCHstart[rank]]][0]), 9*colSize, MPI_DOUBLE, rank, 15, MPI_COMM_WORLD );
    }
  }
  else {
    MPI_Recv( &ARCHmatrixlen, 1, MPI_INT, 0, 13, MPI_COMM_WORLD, &status );
#ifdef DBGDUMP
    fprintf(dbgFile, "Proc receives %d elements of ARCHmatrixcols\n",ARCHmatrixlen);
#endif
    mem_init_K();
    ARCHmatrixcol = (int*)malloc( (ARCHmatrixlen+1)*sizeof(int) );
    MPI_Recv( ARCHmatrixcol, ARCHmatrixlen, MPI_INT, 0, 14, MPI_COMM_WORLD, &status );
    MPI_Recv( &(K[0][0]), 9*ARCHmatrixlen, MPI_DOUBLE, 0, 15, MPI_COMM_WORLD, &status );
  }

  MPI_Barrier(MPI_COMM_WORLD);

  if ( myRank == 0 ) {
    ARCHmatrixlen = ARCHmatrixindex[ARCHstart[1]];
  }
#ifdef DBGDUMP
  fprintf(dbgFile, "Done distributing ARCHmatrixcol arrays\n");
  /* for ( j = 0; j < ARCHmatrixlen; j++ ) */
  /*   fprintf(dbgFile,"col[%d] = %d\n",j,ARCHmatrixcol[j]); */
  fflush(dbgFile);
#endif

  /* GO THROUGH ADJACENCY ARRAY AND MARK MY GHOSTS */
  ghostIndex = (int*)malloc(ARCHglobalnodes*sizeof(int));
  for ( j = 0; j < ARCHglobalnodes; j++ )
    ghostIndex[j] = -1;
  for ( j = 0; j < ARCHmatrixlen; j++ ) {
    if ( (ARCHmatrixcol[j] < ARCHstart[myRank]) ||
	 (ARCHmatrixcol[j] >= ARCHend[myRank]) ) {
      ghostIndex[ARCHmatrixcol[j]] = 1;
    }
  }

#ifdef DBGDUMP
  fprintf(dbgFile, "Done marking ghosts\n");
  fflush(dbgFile);
#endif

  /* CONVERT GHOST MARK TO AN INDEX INTO EXTENDED disp* ARRAYS */
  /* ALSO COUNT NUMBER OF GHOSTS FOR EACH OTHER PROCESS        */
  numGhosts = (int*)malloc(numProcs*sizeof(int));
  index = ARCHnodes;
  for ( rank  = 0; rank < numProcs; rank++ ) {
    numGhosts[rank] = 0;
    if ( rank != myRank ) {
      for ( j = ARCHstart[rank]; j < ARCHend[rank]; j++ ) {
	if ( ghostIndex[j] > 0 ) {
	  ghostIndex[j] = index++;
	  ++numGhosts[rank];
	}
      }
    }
  }

  totalGhosts = index - ARCHnodes;

#ifdef DBGDUMP
  fprintf(dbgFile, "Done indexing ghosts\n");
  fflush(dbgFile);
#endif

  /* EXTEND disp* ARRAYS TO INCLUDE GHOSTS */
  disp0 = (double (*)[3]) realloc( disp0, 3*index*sizeof(double *));
  disp1 = (double (*)[3]) realloc( disp1, 3*index*sizeof(double *));
  disp2 = (double (*)[3]) realloc( disp2, 3*index*sizeof(double *));
  /* disp0[0] = (double*)realloc( disp0[0], 3*index*sizeof(double) ); */
  /* disp1[0] = (double*)realloc( disp1[0], 3*index*sizeof(double) ); */
  /* disp2[0] = (double*)realloc( disp2[0], 3*index*sizeof(double) ); */
  /* for (j = 1; j < index; j++) { */
  /*   disp0[j] = disp0[0] + j*3; */
  /*   disp1[j] = disp1[0] + j*3; */
  /*   disp2[j] = disp2[0] + j*3; */
  /* } */

#ifdef DBGDUMP
  fprintf(dbgFile, "Extended disp arrays from %d to %d\n", ARCHnodes, index);
  fflush(dbgFile);
#endif

  /* STORE START INDICES FOR GHOST ARRAYS */
  ghostStart = (int*)malloc(numProcs*sizeof(int));
  ghostStart[0] = ARCHnodes;
  for ( rank = 1; rank < numProcs; rank++ )
    ghostStart[rank] = ghostStart[rank-1] + numGhosts[rank-1];

  /* STORE THE GLOBAL INDICES OF THE GHOSTS FOR SENDING TO OTHER PROCESS */
  globalGhostIndices = (int**)malloc(numProcs*sizeof(int*));
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numGhosts[rank] > 0 ) {
      index = 0;
      globalGhostIndices[rank] = (int*)malloc(numGhosts[rank]*sizeof(int));
      for ( j = ARCHstart[rank]; j < ARCHend[rank]; j++ ) {
	if ( ghostIndex[j] > 0 )
	  globalGhostIndices[rank][index++] = j;
      }
    }
  }

#ifdef DBGDUMP
  fprintf(dbgFile, "Done storing global ghost indices\n");
  fflush(dbgFile);
#endif
  
  /* ADJUST THE ENTRIES IN ARCHmatrixcol TO BE LOCAL INDICES, INCLUDING GHOSTS */
  for ( j = 0; j < ARCHmatrixlen; j++ ) {
    if ( (ARCHmatrixcol[j] < ARCHstart[myRank]) ||
	 (ARCHmatrixcol[j] >= ARCHend[myRank]) )
      ARCHmatrixcol[j] = ghostIndex[ARCHmatrixcol[j]];
    else
      ARCHmatrixcol[j] -= ARCHstart[myRank];
  }

  MPI_Barrier(MPI_COMM_WORLD);

#ifdef DBGDUMP
  fprintf(dbgFile, "Done adjusting ARCHmatrixcol\n");
  fflush(dbgFile);
#endif

  /*
   *------------------------------------------------------
   * EACH PROCESS HAS THE INFORMATION ABOUT ITS OWN GHOSTS 
   * NOW EXCHANGE INFORMATION SO EACH PROCESS KNOWS WHICH
   * ELEMENTS IT OWNS ARE GHOSTS ON OTHER PROCESSES 
   *------------------------------------------------------
   */
  numRemoteGhosts = (int*)malloc(numProcs*sizeof(int));
  MPI_Alltoall( numGhosts, 1, MPI_INT, numRemoteGhosts, 1, MPI_INT, MPI_COMM_WORLD );

#ifdef DBGDUMP
  fprintf(dbgFile, "Done MPI_Alltoall for numGhosts\n");
  fprintf(dbgFile,"numGhosts:\n");
  for ( rank = 0; rank < numProcs; rank++ ) {
    fprintf(dbgFile,"proc %d: %d\n", rank, numGhosts[rank]);
  }
  fprintf(dbgFile,"\n\nnumRemoteGhosts:\n");
  for ( rank = 0; rank < numProcs; rank++ ) {
    fprintf(dbgFile,"proc %d: %d\n", rank, numRemoteGhosts[rank]);
  }
  fflush(dbgFile);
#endif

  remoteGhostIndices = (int**)malloc(numProcs*sizeof(int*));
  remoteGhostBuffer = (double**)malloc(numProcs*sizeof(double*));
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numRemoteGhosts[rank] > 0 ) {
      remoteGhostIndices[rank] = (int*)malloc(numRemoteGhosts[rank]*sizeof(int));
      remoteGhostBuffer[rank] = (double*)malloc(3*numRemoteGhosts[rank]*sizeof(double));
    }
  }

  /* PACK THE INDICES INTO A SINGLE ARRAY FOR ALLTOALLV */
  packedLGIsize = 0;
  packedRGIsize = 0;
  packedLGIdisplacements = (int*)malloc(numProcs*sizeof(int));
  packedRGIdisplacements = (int*)malloc(numProcs*sizeof(int));
  packedLGIdisplacements[0] = 0;
  packedRGIdisplacements[0] = 0;
  for ( rank = 0; rank < numProcs; rank++ ) {
    /* LOCAL GHOSTS */
    packedLGIsize += numGhosts[rank];
    if ( rank < numProcs-1 )
      packedLGIdisplacements[rank+1] = packedLGIdisplacements[rank] + numGhosts[rank];
    /* REMOTE GHOSTS */
    packedRGIsize += numRemoteGhosts[rank];
    if ( rank < numProcs-1 )
      packedRGIdisplacements[rank+1] = packedRGIdisplacements[rank] + numRemoteGhosts[rank];
  }
  packedLocalGhostIndices = (int*)malloc(packedLGIsize*sizeof(int));
  packedRemoteGhostIndices = (int*)malloc(packedRGIsize*sizeof(int));

  index = 0;
  for ( rank = 0; rank < numProcs; rank++ ) {
    for ( j = 0; j < numGhosts[rank]; j++ ) {
      packedLocalGhostIndices[index++] = globalGhostIndices[rank][j];
    }
  }

  MPI_Alltoallv( packedLocalGhostIndices, numGhosts,
		 packedLGIdisplacements, MPI_INT,
		 packedRemoteGhostIndices, numRemoteGhosts,
		 packedRGIdisplacements, MPI_INT,
		 MPI_COMM_WORLD );

  /* UNPACK REMOTE GHOST INDICES AND ADJUST TO LOCAL INDEX */
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numRemoteGhosts[rank] > 0 ) {
      memcpy( remoteGhostIndices[rank],
	      &packedRemoteGhostIndices[packedRGIdisplacements[rank]],
	      numRemoteGhosts[rank]*sizeof(int) );
      for ( j = 0; j < numRemoteGhosts[rank]; j++ )
	remoteGhostIndices[rank][j] -= ARCHstart[myRank];
    }
  }

  free(packedLocalGhostIndices);
  free(packedRemoteGhostIndices);
  free(packedLGIdisplacements);
  free(packedRGIdisplacements);

  /* determine which process owns sourcenode and epicenter node */
  /* and adjust the index accordingly */
  MPI_Bcast( &(Src.dip), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Src.strike), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Src.rake), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Src.fault), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Src.xyz), 3, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Src.sourcenode), 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &(Src.epicenternode), 1, MPI_INT, 0, MPI_COMM_WORLD );

  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( (Src.sourcenode >= ARCHstart[rank]) &&
	 (Src.sourcenode < ARCHend[rank]) ) {
      sourcenodeRank = rank;
      Src.sourcenode -= ARCHstart[rank];
      break;
    }
  }
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( (Src.epicenternode >= ARCHstart[rank]) &&
	 (Src.epicenternode < ARCHend[rank]) ) {
      epicenternodeRank = rank;
      Src.epicenternode -= ARCHstart[rank];
      break;
    }
  }
}

/*
 *---------------------------------------------------------------
 * add_ghosts_to_disp:
 * each process which has ghost cells of disp[012] sends them to
 * the owning process to be added to the intra-process reduction.
 *---------------------------------------------------------------
 */
void add_ghosts_to_disp( int which_disp )
{
  int j, rank;
  MPI_Status status;
  double *disp_ptr = which_disp == 0 ? disp0[0] : which_disp == 1 ? disp1[0] : disp2[0];

  /* receive remote ghosts from other process and */
  /* move each value to correct position in array */
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numRemoteGhosts[rank] > 0 ) {
      MPI_Recv( remoteGhostBuffer[rank], 3*numRemoteGhosts[rank],
		MPI_DOUBLE, rank, 2, MPI_COMM_WORLD, &status );
      for ( j = 0; j < numRemoteGhosts[rank]; ++j ) {
	disp_ptr[3*remoteGhostIndices[rank][j]]   += remoteGhostBuffer[rank][3*j];
	disp_ptr[3*remoteGhostIndices[rank][j]+1] += remoteGhostBuffer[rank][3*j+1];
	disp_ptr[3*remoteGhostIndices[rank][j]+2] += remoteGhostBuffer[rank][3*j+2];
      }
    }
  }

  /* send ghosts to owning process */
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numGhosts[rank] > 0 ) {
      MPI_Send( disp_ptr + 3*ghostStart[rank], 3*numGhosts[rank], MPI_DOUBLE, rank, 2, MPI_COMM_WORLD );
    }
  }

  /* printf("proc %d done add ghosts to disp\n", myRank); */
  /* fflush(stderr); */

}

/*
 *---------------------------------------------------------------
 * update_disp_ghosts:
 * send updated values to 
 *---------------------------------------------------------------
 */
void update_disp_ghosts( int which_disp )
{
  int j, rank;
  MPI_Status status;
  double *disp_ptr = which_disp == 0 ? disp0[0] : which_disp == 1 ? disp1[0] : disp2[0];

  /* send values from owning proc to remote ghosts */
  /* printf("update disp ghosts SEND\n"); */
  /* fflush(stderr); */
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numRemoteGhosts[rank] > 0 ) {
      /* gather values into buffer */
      /* printf("proc %d sending to proc %d\n", myRank, rank); */
      /* fflush(stderr); */
      for ( j = 0; j < numRemoteGhosts[rank]; ++j ) {
	remoteGhostBuffer[rank][3*j]   = disp_ptr[3*remoteGhostIndices[rank][j]];
	remoteGhostBuffer[rank][3*j+1] = disp_ptr[3*remoteGhostIndices[rank][j]+1];
	remoteGhostBuffer[rank][3*j+2] = disp_ptr[3*remoteGhostIndices[rank][j]+2];
      }
      MPI_Send( remoteGhostBuffer[rank], 3*numRemoteGhosts[rank], MPI_DOUBLE, rank, 1, MPI_COMM_WORLD );
    }
  }

  /* printf("update disp ghosts RECV\n"); */
  /* fflush(stderr); */
  for ( rank = 0; rank < numProcs; rank++ ) {
    if ( numGhosts[rank] > 0 ) {
      /* printf("proc %d recv %d dbls from proc %d at index %d\n", myRank, numGhosts[rank], rank, ghostStart[rank]); */
      /* fflush(stderr); */
      MPI_Recv( disp_ptr + 3*ghostStart[rank], 3*numGhosts[rank], MPI_DOUBLE, rank, 1, MPI_COMM_WORLD, &status );
    }
  }

  /* printf("proc %d done update disp ghosts\n", myRank); */
  /* fflush(stderr); */

}

/*--------------------------------------------------------------------------*/ 
