/*
 * cdefs.cpp: This file is part of the IEC project.
 *
 * IEC : Inspector/Executor Compiler
 *
 * Copyright (C) 2011 the Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P. Sadayappan <saday@cse.ohio-state.edu>
 *
 */
/**
 * @file: cdefs.cpp
 * @author: Mahesh Ravishankar <ravishan@cse.ohio-state.edu>
 */
#include "RunTime/inspector.hpp"
#include "sys/time.h"
#include "mpi.h"
#include "armci.h"

using namespace std;

static double* scalar_holder = NULL;
static double final_val = 0.0;
static ret_data_access data_access_info;
#ifdef _TIMER_
static double inspector_start = 0.0;
static double inspector_stop = 0.0;
static double executor_start = 0.0;
static double executor_stop = 0.0;
#endif

extern "C" {

  void armci_init_(){
    ARMCI_Init();
  }

  void armci_finalize_(){
    ARMCI_Finalize();
  }

  double rtclock() { 
    struct timezone Tzp; 
    struct timeval Tp; 
    int stat; 
    stat = gettimeofday (&Tp, &Tzp); 
    if (stat != 0) printf("Error return from gettimeofday: %d",stat); 
    return(Tp.tv_sec + Tp.tv_usec*1.0e-6); 
  }


  void create_inspector(int md, int np, int nt, int niters, int ndata, int nc, int nac, int* nic, int* ndc, int* ro){
#ifdef _TIMER_
    inspector_start = rtclock();
#endif
    inspector* new_inspector = inspector::instance(md,np,nt,niters,ndata,nc,nac,nic,ndc,ro);
    scalar_holder = new double[nt];
#ifndef NDEBUG
    fprintf(stdout,"PID:%d,Creating inspector:%p\n",md,new_inspector);
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
#endif
  }

  void create_inspector_(int* md, int* np, int* nt, int* niters, int* ndata, int* nc, int *nac, int *nic, int* ndc, int* ro){
    create_inspector(*md,*np,*nt,*niters,*ndata,*nc,*nac,nic,ndc,ro);
  }


  void set_access_array_param(int an,int as,int st,int* oa){
    inspector::instance()->SetAccessArrayParam(an,as,st,oa);
  }

  void set_access_array_param_(int* an, int *as, int* st, int *oa){
    set_access_array_param(*an,*as,*st,oa);
  }
  

  int is_known(int an, int indx){
    if( inspector::instance()->HaveIndex(an,indx) )
      return 1;
    else
      return 0;
  }

  void is_known_(int* an, int* indx, int* val){
    *val = is_known(*an,*indx);
  }

  int get_elem(int an, int index){
    return inspector::instance()->GetIndex(an,index);
  }

  void get_elem_(int* an, int* index, int* val){
    *val = get_elem(*an,*index);
  }

  int done_graph_gen(){
    if( inspector::instance()->DoneGraphGeneration() )
      return 1;
    else
      return 0;
  }

  void done_graph_gen_(int* val){
    *val = done_graph_gen();
  }
  
  void add_vertex(int iter_num, int iter_value) {
    inspector* my_inspect = inspector::instance();
    my_inspect->AddVertex(iter_num,iter_value);
  }

  
  void add_vertex_(int* iter_num, int* iter_value){
    inspector* my_inspect = inspector::instance();
    my_inspect->AddVertex(*iter_num,*iter_value);
  }


  void add_pin_to_net(int data_num, int index, int id, int isploop) {
    //  printf("Adding pin to %d of %d\n",index,data_num);
    inspector* my_inspect = inspector::instance();
    my_inspect->AddNet(data_num,index,id,isploop);
  }


  void add_pin_to_net_(int* data_num, int* index, int* id, int* isploop){
    inspector* my_inspect = inspector::instance();
    my_inspect->AddNet(*data_num,*index,*id,*isploop);
  }

  void partition_hypergraph(int use_type)  {
    switch(use_type){
    case 0:
      inspector::instance()->PatohPartition();
      return;
    case 1:
      inspector::instance()->MetisPartition();
      return;
    case 2:
      inspector::instance()->BlockPartition();
      return;
    default:
      assert(0);
    }
  }

  void partition_hypergraph_(int *type){
    partition_hypergraph(*type);
  }

  
  void print_hypergraph() {  
#ifndef NDEBUG
    printf("ID:%d,Here:\n",inspector::instance()->get_proc_id());
    fflush(stdout);
    char graph_file[18];
    sprintf(graph_file,"hypergraph_%d.dat",inspector::instance()->get_proc_id());
    FILE* outfile = fopen(graph_file,"w");
    inspector::instance()->print_hypergraph(outfile);
    fclose(outfile);
#endif
  }

  void print_hypergraph_(){
    print_hypergraph();
  }
  
  int get_proc_iter_size(int in){
    return inspector::instance()->GetProcLocal(in);
  }

  void get_proc_iter_size_(int *in, int *out){
    *out = inspector::instance()->GetProcLocal(*in);
  }

  void set_array_stride(int an, int st) {
    inspector::instance()->SetStride(an,st);
  }


  void set_array_stride_(int* an, int* st) {
    inspector::instance()->SetStride(*an,*st);
  }


  int get_vertex_home(int in ,int iv){
    return inspector::instance()->GetVertexHome(in,iv);
  }


  void get_vertex_home_(int* in ,int* iv, int* out){
    *out = inspector::instance()->GetVertexHome(*in,*iv);
  }

  int get_local_size(int thread_id, int an){
    if( thread_id == 0 ){
      data_access_info = inspector::instance()->GetLocalAccesses(an);
    }
#pragma omp barrier
    int * proc_recv_buffer = data_access_info.recvbuffer;
    int * proc_recv_displ = data_access_info.recvdispl;
    int * proc_recv_count = data_access_info.recvcount;
    local_inspector* curr_local_inspector = local_inspector::instance(thread_id);
    int nprocs = inspector::instance()->GetNProcs();
    for( int j = 0 ; j < nprocs ; j++ ){
      curr_local_inspector->InsertDirectAccess(an,proc_recv_buffer+proc_recv_displ[thread_id*nprocs*2+j*2],proc_recv_count[thread_id*nprocs*2+j*2]);
      curr_local_inspector->InsertIndirectAccess(an,proc_recv_buffer+proc_recv_displ[thread_id*nprocs*2+j*2+1],proc_recv_count[thread_id*nprocs*2+j*2+1]);
    }
#pragma omp barrier 
    if( thread_id == 0 ){
      delete[] data_access_info.recvbuffer;
      delete[] data_access_info.recvcount;
      delete[] data_access_info.recvdispl;
    }
    curr_local_inspector->SetupLocalArray(an);
   return curr_local_inspector->GetLocalDataSize(an);
  }


  void get_local_size_(int* thread_id, int* an, int* out){
    *out = get_local_size(*thread_id,*an);
  }
  
  void communicate_reads_for(int thread_id, int in, int an){
    local_inspector::instance(thread_id)->AddReadArray(in,an);
  }


  void communicate_reads_for_(int *thread_id, int *in, int *an){
    local_inspector::instance(*thread_id)->AddReadArray(*in,*an);
  }

  void communicate_writes_for(int thread_id, int in, int an){
    local_inspector::instance(thread_id)->AddWriteArray(in,an);
  }

  void communicate_writes_for_(int *thread_id, int *in, int *an){
    local_inspector::instance(*thread_id)->AddWriteArray(*in,*an);
  }

  void add_index_from_proc(int thread_id, int data_num, int index, int access_type){
    local_inspector::instance(thread_id)->AddIndexAccessed(data_num,index,access_type);
  }

  void add_index_from_proc_(int *thread_id, int *data_num, int *index, int *access_type){
    local_inspector::instance(*thread_id)->AddIndexAccessed(*data_num,*index,*access_type);
  }


  void populate_local_array(int thread_id, int an, double*lb, double*oa, int st){
    local_inspector::instance(thread_id)->PopulateLocalArray(an,lb,oa,st);
  }

  void populate_local_array_(int *thread_id, int *an, double*lb, double*oa, int *st){
    local_inspector::instance(*thread_id)->PopulateLocalArray(*an,lb,oa,*st);
  }


  void renumber_access_array(int thread_id, int an, int as, int* aa){
    local_inspector::instance(thread_id)->RenumberAccessArray(an,as,aa);
  }

  void renumber_access_array_(int *thread_id, int *an, int *as, int* aa){
    local_inspector::instance(*thread_id)->RenumberAccessArray(*an,*as,aa);
  }

  void renumber_offset_array(int thread_id, int an, int as, int* aa, int* la){
    local_inspector::instance(thread_id)->RenumberOffsetArray(an,as,aa,la);
  }

  void renumber_offset_array_(int *thread_id, int *an, int *as, int* aa, int* la){
    local_inspector::instance(*thread_id)->RenumberOffsetArray(*an,*as,aa,la);
  }

  void setup_executor(int thread_id){
#ifndef NDEBUG
    printf("ID = %d, Going to call setup executor (%p)\n",thread_id,local_inspector::instance(thread_id));
    fflush(stdout);
#endif
    local_inspector::instance(thread_id)->GenerateGhosts();
    
#ifndef NDEBUG
    printf("LocalGhostsDone\n");
    fflush(stdout);
#endif

#pragma omp barrier
    
    if( thread_id == 0 ){
      inspector::instance()->CommunicateGhosts();
#ifndef NDEBUG
      printf("Global ghosts done\n");
      fflush(stdout);
#endif
    }
#pragma omp barrier

    local_inspector::instance(thread_id)->GenerateOwned();
    

#pragma omp barrier

    if( thread_id == 0 ){
      inspector::instance()->GetBufferSize();
#ifdef _TIMER_
      inspector_stop = rtclock();
      if( inspector::instance()->get_proc_id() == 0 )
	printf("[IEC]:InspectorTime:%lf\n",inspector_stop-inspector_start);
#endif
    }

#pragma omp barrier

#ifdef _TIMER_
    if( thread_id == 0 ){
      executor_start = rtclock();
    }
#endif
  }

  
  void setup_executor_(int * thread_id){
    setup_executor(*thread_id);
  }

  void communicate_reads(int tid, int cn){
    inspector::instance()->CommunicateReads(tid,cn);
  }

  void communicate_reads_(int *tid, int *cn){
    inspector::instance()->CommunicateReads(*tid,*cn);
  }


  void communicate_writes(int tid, int cn){
    inspector::instance()->CommunicateWrites(tid,cn);
  }


  void communicate_writes_(int *tid, int *cn){
    inspector::instance()->CommunicateWrites(*tid,*cn);
  }

  void init_write_ghosts(int tid, int cn){
    local_inspector::instance(tid)->InitWriteGhosts(cn);
  }

  void init_write_ghosts_(int *tid, int *cn){
    local_inspector::instance(*tid)->InitWriteGhosts(*cn);
  }

  void initialize_scalar(int proc_id, int thread_id, double * val){
    if( thread_id != 0 || proc_id != 0 )
      *val = 0.0;
  }

  void initialize_scalar_(int * proc_id , int * thread_id, double * val ){
    initialize_scalar(*proc_id,*thread_id,val);
  }

  void reduce_scalar(int thread_id, double *val){
    scalar_holder[thread_id] = *val;
#pragma omp barrier

#pragma omp master
    {
      int nthreads = inspector::instance()->GetNThreads();
      final_val = 0.0;
      for( int i =0  ; i < nthreads ; i++ )
	final_val += scalar_holder[i];
      
      MPI_Allreduce(MPI_IN_PLACE,&final_val,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
    }

#pragma omp barrier
    *val = final_val;
  }

  void reduce_scalar_(int* thread_id, double* val){
    reduce_scalar(*thread_id,val);
  }

  void populate_global_arrays(int tid){
    if( tid == 0 ){
#ifdef _TIMER_
      executor_stop = rtclock();
      if( inspector::instance()->get_proc_id() == 0 )
	printf("[IEC]:ExecutorTime:%lf\n",executor_stop-executor_start);
#endif
    }
 
    local_inspector::instance(tid)->PopulateGlobalArrays();
  }

  void populate_global_arrays_(int *tid){
    local_inspector::instance(*tid)->PopulateGlobalArrays();
  }


  void delete_inspector(int tid){
    delete local_inspector::instance(tid);

#pragma omp barrier

#pragma omp master
    {
      delete inspector::instance();
      delete[] scalar_holder;
    }

#pragma omp barrier
  }

  void delete_inspector_(int* tid){
    delete_inspector(*tid);
  }

  void print_data(int thread_id){
#ifndef NDEBUG
    local_inspector::instance(thread_id)->print_data();
    if( thread_id == 0 )
      inspector::instance()->print_comm();
#endif
  }

  void print_data_(int* thread_id){
    print_data(*thread_id);
  }


  void print_access(){
#ifndef NDEBUG
    printf("PID:%d,PrinitingAccess\n",inspector::instance()->get_proc_id());
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
    inspector::instance()->print_access();
#endif
  }

  void print_access_(){
    print_access();
  }

  
  //dim1 = number of rows, dim2 number of columns. Row major (colums is fastest varying)
  double** malloc_2d_double(int dim1, int dim2){
    double* base_array = (double*)malloc(dim1*dim2*sizeof(double));
    double** actual_array = (double**)malloc(dim1*sizeof(double*));
    for( int i = 0 ; i < dim1 ; i++ ){
      actual_array[i] = base_array + i*dim2;
    }
    return actual_array;
  }

  float** malloc_2d_float(int dim1, int dim2){
    float* base_array = (float*)malloc(dim1*dim2*sizeof(float));
    float** actual_array = (float**)malloc(dim1*sizeof(float*));
    for( int i = 0 ; i < dim1 ; i++ ){
      actual_array[i] = base_array + i*dim2;
    }
    return actual_array;
  }

  void free_2d_double(double** actual_array){
    free(actual_array[0]);
    free(actual_array);
  }

  void free_2d_float(float** actual_array){
    free(actual_array[0]);
    free(actual_array);
  }


  int  init_solver(int s){
    return inspector::instance()->InitSolver(s);
  }

  void init_solver_(int *s, int *sn){
    *sn = inspector::instance()->InitSolver(*s);
  }
  
  void add_unknown(int sn, int an, int idx, int rn){
    assert(sn == 0);
    inspector::instance()->AddUnknown(sn,an,idx,rn);
  }

  void add_unknown_(int *sn, int *an, int *idx, int *rn){
    assert(*sn == 0);
    inspector::instance()->AddUnknown(*sn,*an,*idx,*rn);
  }

  void renumber_solver_rows(int sn, int* oa, int as){
    assert(sn == 0 );
    inspector::instance()->RenumberGlobalRows(sn,oa,as);
  }

  void renumber_solver_rows_(int *sn, int* oa, int *as){
    assert(*sn == 0 );
    inspector::instance()->RenumberGlobalRows(*sn,oa,*as);
  }

  void print_solver(){
    inspector::instance()->print_solver();
  }

  void print_solver_(){
    inspector::instance()->print_solver();
  }

  int get_local_rows(int sn, int tid){
    return inspector::instance()->GetLocalRows(sn,tid);
  }

  void get_local_rows_(int *sn, int *tid, int *lr){
    *lr =  inspector::instance()->GetLocalRows(*sn,*tid);
  }

  void set_constraint(int an){
    inspector::instance()->SetConstraint(an);
  }

  void set_constraint_(int *an){
    inspector::instance()->SetConstraint(*an);
  }

  void set_local_array(int tid, int an, void* la){
    local_inspector::instance(tid)->SetLocalArray(an,la);
  }

  void set_local_array_(int* tid, int* an, void* la){
    local_inspector::instance(*tid)->SetLocalArray(*an,la);
  }

  int get_local_block_offset(int tid, int an){
    return local_inspector::instance(tid)->GetLocalBlockOffset(an);
  }

  void get_local_block_offset_(int* tid, int* an, int *bo){
    *bo = local_inspector::instance(*tid)->GetLocalBlockOffset(*an);
  }

  void fflush_(){
    fflush(stdout);
    fflush(stderr);
  }
  
}
